r"""Continuous linear change"""
import numpy as np

from ruptures.base import BaseCost
from ruptures.costs import NotEnoughPoints


class CostCLinear(BaseCost):

    r"""Piecewise linear approximation with a continuity constraint."""

    model = "clinear"

    def __init__(self):
        """Initialize the object."""
        self.signal = None
        self.min_size = 3

    def fit(self, signal) -> "CostCLinear":
        """Set parameters of the instance.

        Args:
            signal (array): signal of shape (n_samples, n_dims) or (n_samples,)

        Returns:
            self
        """
        if signal.ndim == 1:
            self.signal = signal.reshape(-1, 1)
        else:
            self.signal = signal

        return self

    def error(self, start, end) -> float:
        """Return the approximation cost on the segment [start:end].

        Args:
            start (int): start of the segment
            end (int): end of the segment

        Returns:
            segment cost (float)

        Raises:
            NotEnoughPoints: when the segment is too short (less than `min_size`
                samples).
        """
        if end - start < self.min_size:
            raise NotEnoughPoints

        if start == 0:
            start = 1

        sub = self.signal[start:end]
        slope = (self.signal[end - 1] - self.signal[start - 1]) / (end - start)
        intercept = self.signal[start - 1]
        approx = slope.reshape(-1, 1) * np.arange(
            1, end - start + 1
        ) + intercept.reshape(-1, 1)
        return np.sum((sub - approx.transpose()) ** 2)



"""The :mod:`ruptures.exceptions` module includes all custom warnings and error
classes used across ruptures."""


class NotEnoughPoints(Exception):

    """Raise this exception when there is not enough point to calculate a cost
    function."""

    pass



r"""Shift in linear model"""
import numpy as np
from numpy.random import normal

from . import pw_constant
from ruptures.utils import draw_bkps


def pw_linear(n_samples=200, n_features=1, n_bkps=3, noise_std=None):
    """Return piecewise linear signal and the associated changepoints.

    Args:
        n_samples (int, optional): signal length
        n_features (int, optional): number of covariates
        n_bkps (int, optional): number of change points
        noise_std (float, optional): noise std. If None, no noise is added
    Returns:
        tuple: signal of shape (n_samples, n_features+1), list of breakpoints
    """

    covar = normal(size=(n_samples, n_features))
    linear_coeff, bkps = pw_constant(
        n_samples=n_samples, n_bkps=n_bkps, n_features=n_features, noise_std=None
    )
    var = np.sum(linear_coeff * covar, axis=1)
    if noise_std is not None:
        var += normal(scale=noise_std, size=var.shape)
    signal = np.c_[var, covar]
    return signal, bkps



cimport ruptures.utils._utils.convert_path_matrix as convert_pm

cimport cython
import numpy as np

cpdef from_path_matrix_to_bkps_list(int[:] path_matrix_flat, int n_bkps, int n_samples, int n_bkps_max, int jump):
    # Init bkps_list array
    cdef int[::1] bkps_list = np.empty((n_bkps+1), dtype=np.dtype("i"))
    try:
        convert_pm.convert_path_matrix_c(&path_matrix_flat[0], n_bkps, n_samples, n_bkps_max, jump, &bkps_list[0])
    except:
        print("An exception occurred.")
    return np.asarray(bkps_list).tolist()


from itertools import product

import numpy as np
import pytest

from ruptures.costs import NotEnoughPoints, CostAR
from ruptures.datasets import pw_constant
from ruptures.detection import Binseg, BottomUp, Dynp, Pelt, Window, KernelCPD


@pytest.fixture(scope="module")
def signal_bkps_5D():
    signal, bkps = pw_constant(n_features=5, noise_std=1)
    return signal, bkps


@pytest.fixture(scope="module")
def signal_bkps_1D():
    signal, bkps = pw_constant(noise_std=1)
    return signal.astype(np.float32), bkps


@pytest.fixture(scope="module")
def signal_bkps_5D_no_noise():
    signal, bkps = pw_constant(n_features=5, noise_std=0)
    return signal, bkps


@pytest.fixture(scope="module")
def signal_bkps_1D_no_noise():
    signal, bkps = pw_constant(noise_std=0)
    return signal, bkps


@pytest.fixture(scope="module")
def signal_bkps_1D_constant():
    signal, bkps = np.zeros(200), [200]
    return signal, bkps


@pytest.mark.parametrize("algo", [Binseg, BottomUp, Dynp, Pelt, Window])
def test_empty(signal_bkps_1D, algo):
    signal, _ = signal_bkps_1D
    algo().fit(signal).predict(1)
    algo().fit_predict(signal, 1)


@pytest.mark.parametrize(
    "algo, model",
    product([Binseg, BottomUp, Window], ["l1", "l2", "ar", "normal", "rbf", "rank"]),
)
def test_model_1D(signal_bkps_1D, algo, model):
    signal, _ = signal_bkps_1D
    algo(model=model).fit_predict(signal, pen=1)
    ret = algo(model=model).fit_predict(signal, n_bkps=1)
    assert len(ret) == 2
    assert ret[-1] == signal.shape[0]
    algo(model=model).fit_predict(signal, epsilon=10)


@pytest.mark.parametrize(
    "algo, model", product([Dynp, Pelt], ["l1", "l2", "ar", "normal", "rbf", "rank"])
)
def test_model_1D_bis(signal_bkps_1D, algo, model):
    signal, bkps = signal_bkps_1D
    algo_t = algo(model=model)
    ret = algo_t.fit_predict(signal, 1)
    if isinstance(algo_t, Dynp):
        assert len(ret) == 2
    assert ret[-1] == signal.shape[0]


@pytest.mark.parametrize(
    "algo, model",
    product(
        [Dynp, Pelt, Binseg, BottomUp, Window],
        ["l1", "l2", "ar", "normal", "rbf", "rank"],
    ),
)
def test_model_1D_constant(signal_bkps_1D_constant, algo, model):
    signal, bkps = signal_bkps_1D_constant
    algo_t = algo(model=model)
    ret = algo_t.fit_predict(signal, 1)
    if isinstance(algo_t, Dynp) or isinstance(algo_t, BottomUp):
        # With constant signal, those search methods
        # will return another break points alongside signal.shape[0]
        assert len(ret) == 2
    if isinstance(algo_t, Binseg):
        if model == "normal":
            # With constant signal, this search method with normal cost
            # will return only signal.shape[0] as breaking points
            assert len(ret) == 1
        else:
            # With constant signal, this search method with another cost
            # will return another break points alongside signal.shape[0]
            assert len(ret) == 2
    if isinstance(algo_t, Window):
        # With constant signal, this search methods
        # will return only signal.shape[0] as breaking points
        assert len(ret) == 1
    if isinstance(algo_t, Pelt):
        assert len(ret) <= 2
    assert ret[-1] == signal.shape[0]


@pytest.mark.parametrize(
    "algo, model",
    product(
        [Binseg, BottomUp, Window], ["l1", "l2", "linear", "normal", "rbf", "rank"]
    ),
)
def test_model_5D(signal_bkps_5D, algo, model):
    signal, _ = signal_bkps_5D
    algo(model=model).fit_predict(signal, pen=1)
    ret = algo(model=model).fit_predict(signal, n_bkps=1)
    assert len(ret) == 2
    algo(model=model).fit_predict(signal, epsilon=10)


@pytest.mark.parametrize(
    "algo, model",
    product([Dynp, Pelt], ["l1", "l2", "linear", "normal", "rbf", "rank"]),
)
def test_model_5D_bis(signal_bkps_5D, algo, model):
    signal, _ = signal_bkps_5D
    algo_t = algo(model=model)
    ret = algo_t.fit_predict(signal, 1)
    if isinstance(algo_t, Dynp):
        assert len(ret) == 2


@pytest.mark.parametrize("algo", [Binseg, BottomUp, Window, Dynp, Pelt])
def test_custom_cost(signal_bkps_1D, algo):
    signal, _ = signal_bkps_1D
    c = CostAR(order=10)
    algo_t = algo(custom_cost=c)
    ret = algo_t.fit_predict(signal, 1)
    if isinstance(algo_t, Pelt):
        assert len(ret) >= 2
    else:
        assert len(ret) == 2


@pytest.mark.parametrize("algo", [Binseg, BottomUp, Window, Dynp, Pelt])
def test_pass_param_to_cost(signal_bkps_1D, algo):
    signal, _ = signal_bkps_1D
    algo_t = algo(model="ar", params={"order": 10})
    ret = algo_t.fit_predict(signal, 1)
    if isinstance(algo_t, Pelt):
        assert len(ret) >= 2
    else:
        assert len(ret) == 2


@pytest.mark.parametrize(
    "algo, kernel, min_size",
    product([KernelCPD], ["linear"], [2, 5]),
)
def test_cython_dynp_1D_linear(signal_bkps_1D, algo, kernel, min_size):
    signal, bkps = signal_bkps_1D
    ret = (
        algo(kernel=kernel, min_size=min_size, jump=1)
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert len(ret) == len(bkps)


@pytest.mark.parametrize(
    "algo, kernel, min_size",
    product([KernelCPD], ["linear"], [2, 5]),
)
def test_cython_dynp_5D_linear(signal_bkps_5D, algo, kernel, min_size):
    signal, bkps = signal_bkps_5D
    ret = (
        algo(kernel=kernel, min_size=min_size, jump=1)
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert len(ret) == len(bkps)


@pytest.mark.parametrize(
    "algo, kernel, min_size",
    product([KernelCPD], ["rbf"], [2, 5]),
)
def test_cython_dynp_1D_rbf(signal_bkps_1D, algo, kernel, min_size):
    signal, bkps = signal_bkps_1D
    ret = (
        algo(kernel=kernel, min_size=min_size, jump=1, params={"gamma": 1.5})
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert len(ret) == len(bkps)


@pytest.mark.parametrize(
    "algo, kernel, min_size",
    product([KernelCPD], ["rbf"], [2, 5]),
)
def test_cython_dynp_5D_rbf(signal_bkps_5D, algo, kernel, min_size):
    signal, bkps = signal_bkps_5D
    ret = (
        algo(kernel=kernel, min_size=min_size, jump=1, params={"gamma": 1.5})
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert len(ret) == len(bkps)


@pytest.mark.parametrize(
    "algo, kernel, min_size",
    product([KernelCPD], ["linear"], [2, 5]),
)
def test_cython_dynp_1D_no_noise_linear(
    signal_bkps_1D_no_noise, algo, kernel, min_size
):
    signal, bkps = signal_bkps_1D_no_noise
    res = (
        algo(kernel=kernel, min_size=min_size, jump=1)
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert res == bkps


@pytest.mark.parametrize(
    "algo, kernel, min_size",
    product([KernelCPD], ["linear"], [2, 5]),
)
def test_cython_dynp_5D_no_noise_linear(
    signal_bkps_5D_no_noise, algo, kernel, min_size
):
    signal, bkps = signal_bkps_5D_no_noise
    res = (
        algo(kernel=kernel, min_size=min_size, jump=1)
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert res == bkps


@pytest.mark.parametrize(
    "algo, kernel, min_size",
    product([KernelCPD], ["rbf"], [2, 5]),
)
def test_cython_dynp_1D_no_noise_rbf(signal_bkps_1D_no_noise, algo, kernel, min_size):
    signal, bkps = signal_bkps_1D_no_noise
    res = (
        algo(kernel=kernel, min_size=min_size, jump=1, params={"gamma": 1.5})
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert res == bkps


@pytest.mark.parametrize(
    "algo, kernel, min_size",
    product([KernelCPD], ["rbf"], [2, 5]),
)
def test_cython_dynp_5D_no_noise_rbf(signal_bkps_5D_no_noise, algo, kernel, min_size):
    signal, bkps = signal_bkps_5D_no_noise
    res = (
        algo(kernel=kernel, min_size=min_size, jump=1, params={"gamma": 1.5})
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert res == bkps


# Exhaustive test of KernelCPD
@pytest.mark.parametrize(
    "algo, kernel",
    product([KernelCPD], ["linear", "rbf", "cosine"]),
)
def test_kernelcpd(signal_bkps_5D, algo, kernel):
    signal, bkps = signal_bkps_5D
    # Test we do not compute if intermediary results exist
    algo_temp = algo(kernel=kernel)
    algo_temp.fit(signal).predict(n_bkps=len(bkps) - 1)
    algo_temp.predict(n_bkps=1)
    # Test penalized version
    algo(kernel=kernel).fit(signal).predict(pen=0.2)
    # Test fit_predict
    algo(kernel=kernel).fit_predict(signal, pen=0.2)



import pytest

from ruptures.datasets import pw_constant
from ruptures.show import display
from ruptures.show.display import MatplotlibMissingError


@pytest.fixture(scope="module")
def signal_bkps():
    signal, bkps = pw_constant()
    return signal, bkps


def test_display_with_options(signal_bkps):
    try:
        signal, bkps = signal_bkps
        fig, axarr = display(signal, bkps)
        fig, axarr = display(signal, bkps, bkps)
        figsize = (20, 10)  # figure size
        fig, axarr = display(
            signal,
            bkps,
            figsize=figsize,
        )
        fig, axarr = display(
            signal[:, 0],
            bkps,
            figsize=figsize,
        )
    except MatplotlibMissingError:
        pytest.skip("matplotlib is not installed")


def test_display_without_options(signal_bkps):
    try:
        signal, bkps = signal_bkps
        fig, axarr = display(signal, bkps)
        fig, axarr = display(signal, bkps, bkps)
        figsize = (20, 10)  # figure size
        fig, axarr = display(signal, bkps)
        fig, axarr = display(signal[:, 0], bkps)
    except MatplotlibMissingError:
        pytest.skip("matplotlib is not installed")


def test_display_with_new_options(signal_bkps):
    try:
        signal, bkps = signal_bkps
        fig, axarr = display(signal, bkps)
        fig, axarr = display(signal, bkps, bkps)

        fig, axarr = display(signal, bkps, facecolor="k", edgecolor="b")
        fig, axarr = display(signal[:, 0], bkps, facecolor="k", edgecolor="b")
    except MatplotlibMissingError:
        pytest.skip("matplotlib is not installed")



r"""Linear model change"""
from numpy.linalg import lstsq

from ruptures.base import BaseCost
from ruptures.costs import NotEnoughPoints


class CostLinear(BaseCost):

    r"""Least-square estimate for linear changes."""

    model = "linear"

    def __init__(self):
        """Initialize the object."""
        self.signal = None
        self.covar = None
        self.min_size = 2

    def fit(self, signal) -> "CostLinear":
        """Set parameters of the instance. The first column contains the
        observed variable. The other columns contains the covariates.

        Args:
            signal (array): signal of shape (n_samples, n_regressors+1)

        Returns:
            self
        """
        assert signal.ndim > 1, "Not enough dimensions"

        self.signal = signal[:, 0].reshape(-1, 1)
        self.covar = signal[:, 1:]
        return self

    def error(self, start, end) -> float:
        """Return the approximation cost on the segment [start:end].

        Args:
            start (int): start of the segment
            end (int): end of the segment

        Returns:
            segment cost

        Raises:
            NotEnoughPoints: when the segment is too short (less than `min_size` samples).
        """
        if end - start < self.min_size:
            raise NotEnoughPoints
        y, X = self.signal[start:end], self.covar[start:end]
        _, residual, _, _ = lstsq(X, y, rcond=None)
        return residual.sum()
